#include "ui.h"
#include "global.h"
#include <cstdio>
#include <lvgl.h>
static lv_obj_t *spinbox_pendulum_length;
static lv_obj_t *spinbox_radius;
static lv_obj_t *spinbox_angle_in_degree;
static lv_obj_t *label_zero_info;
static lv_obj_t *label_imu_display;

static lv_obj_t *spinbox_pitch_p;
static lv_obj_t *spinbox_pitch_i;
static lv_obj_t *spinbox_pitch_d;
static lv_obj_t *spinbox_roll_p;
static lv_obj_t *spinbox_roll_i;
static lv_obj_t *spinbox_roll_d;

static void btn_zero_on_clicked(lv_event_t *)
{
    constexpr const int times = 100;
    constexpr const int delay_ticks = 10;
    int32_t pitch_sum = 0;
    int32_t roll_sum = 0;
    for (int i = 0; i < times; ++i)
    {
        pitch_sum += jy61_device.attitude_raw.pitch_raw;
        roll_sum += jy61_device.attitude_raw.roll_raw;
        HAL_Delay(delay_ticks);
    }
    pitch_bias_raw = static_cast<int16_t>(pitch_sum / times);
    roll_bias_raw = static_cast<int16_t>(roll_sum / times);
}

static void btn_off_on_clicked(lv_event_t *)
{
    fan_output = false;
}

static void btn_brake_on_clicked(lv_event_t *)
{
    fan_output = true;
    time_elapsed = -3;
    curve_pitch.amplitude = curve_roll.amplitude = 0;
    curve_pitch.angular_frequency = curve_roll.angular_frequency = 0;
    curve_pitch.phase = curve_roll.phase = 0;
}

static void btn_line_on_clicked(lv_event_t *)
{
    fan_output = true;
    time_elapsed = -3;

    double angle_in_degree = lv_spinbox_get_value(spinbox_angle_in_degree);
    double angle = angle_in_degree * M_PI / 180.0;
    double pendulum_length = lv_spinbox_get_value(spinbox_pendulum_length);
    double radius = lv_spinbox_get_value(spinbox_radius);
    curve_pitch.amplitude = static_cast<float>(asin(radius * cos(angle) / pendulum_length));
    curve_roll.amplitude = static_cast<float>(asin(radius * sin(angle) / pendulum_length));
    curve_pitch.angular_frequency = curve_roll.angular_frequency = 3.54;
    curve_pitch.phase = curve_roll.phase = M_PI_2;
}

static void btn_circle_on_clicked(lv_event_t *)
{
    fan_output = true;
    time_elapsed = -3;

    double pendulum_length = lv_spinbox_get_value(spinbox_pendulum_length);
    double radius = lv_spinbox_get_value(spinbox_radius);
    curve_pitch.amplitude = curve_roll.amplitude = static_cast<float>(asin(radius / pendulum_length));
    curve_pitch.angular_frequency = curve_roll.angular_frequency = 3.54;
    curve_pitch.phase = 0;
    curve_roll.phase = M_PI_2;
}

static void spinbox_pitch_p_on_changed(lv_event_t *)
{
    pid_pitch.Kp = (float)(lv_spinbox_get_value(spinbox_pitch_p) / 1.0e3);
}
static void spinbox_pitch_i_on_changed(lv_event_t *)
{
    pid_pitch.Ki = (float)(lv_spinbox_get_value(spinbox_pitch_i) / 1.0e3);
}
static void spinbox_pitch_d_on_changed(lv_event_t *)
{
    pid_pitch.Kd = (float)(lv_spinbox_get_value(spinbox_pitch_d) / 1.0e3);
}
static void spinbox_roll_p_on_changed(lv_event_t *)
{
    pid_roll.Kp = (float)(lv_spinbox_get_value(spinbox_roll_p) / 1.0e3);
}
static void spinbox_roll_i_on_changed(lv_event_t *)
{
    pid_roll.Ki = (float)(lv_spinbox_get_value(spinbox_roll_i) / 1.0e3);
}
static void spinbox_roll_d_on_changed(lv_event_t *)
{
    pid_roll.Kd = (float)(lv_spinbox_get_value(spinbox_roll_d) / 1.0e3);
}

static void timer_update_imu_display(lv_timer_t *)
{
    char repr[256];
    double pitch = (jy61_device.attitude_raw.pitch_raw - pitch_bias_raw) / 32768.0 * 180;
    double roll = (jy61_device.attitude_raw.roll_raw - roll_bias_raw) / 32768.0 * 180;
    double yaw = jy61_device.attitude_raw.yaw_raw / 32768.0 * 180;
    sprintf(repr, "IMU(Degree): P=%.1lf,R=%.1lf,Y=%.1lf", pitch, roll, yaw);
    lv_label_set_text(label_imu_display, repr);

    sprintf(repr, "Bias: P=%d,R=%d", pitch_bias_raw, roll_bias_raw);
    lv_label_set_text(label_zero_info, repr);
}

static lv_obj_t *add_spinbox_row(lv_group_t *g, lv_obj_t *cont, uint8_t row_pos, const char *label_text,
                                 int32_t range_min, int32_t range_max, uint8_t digit_count, uint8_t separator_position)
{
    lv_obj_t *label = lv_label_create(cont);
    lv_label_set_text(label, label_text);
    lv_obj_set_grid_cell(label, LV_GRID_ALIGN_START, 0, 1, LV_GRID_ALIGN_CENTER, row_pos, 1);
    lv_obj_t *spinbox = lv_spinbox_create(cont);
    lv_spinbox_set_range(spinbox, range_min, range_max);
    lv_spinbox_set_digit_format(spinbox, digit_count, separator_position);
    lv_spinbox_set_value(spinbox, 0);
    lv_spinbox_set_step(spinbox, 1);
    lv_obj_set_grid_cell(spinbox, LV_GRID_ALIGN_START, 1, 1, LV_GRID_ALIGN_CENTER, row_pos, 1);
    lv_group_add_obj(g, spinbox);
    return spinbox;
}

static lv_obj_t *add_button_cell(lv_group_t *g, lv_obj_t *cont, uint8_t row_pos, uint8_t col_pos,
                                 const char *label_text)
{
    lv_obj_t *btn = lv_btn_create(cont);
    lv_obj_set_grid_cell(btn, LV_GRID_ALIGN_STRETCH, col_pos, 1, LV_GRID_ALIGN_CENTER, row_pos, 1);
    lv_group_add_obj(g, btn);
    lv_obj_t *label = lv_label_create(btn);
    lv_label_set_text(label, label_text);
    lv_obj_center(label);
    return btn;
}

void ui_init(lv_indev_t *indev)
{
    lv_group_t *g = lv_group_create();
    lv_indev_set_group(indev, g);

    static lv_coord_t col_dsc[] = {90, 90, LV_GRID_TEMPLATE_LAST};
    static lv_coord_t row_dsc[] = {48, 48, 48, 48, 48, 48, 48, 48, 48, 48, 48, 48, 48, LV_GRID_TEMPLATE_LAST};

    lv_obj_t *cont = lv_obj_create(lv_scr_act());
    lv_obj_set_style_grid_column_dsc_array(cont, col_dsc, 0);
    lv_obj_set_style_grid_row_dsc_array(cont, row_dsc, 0);
    lv_obj_set_size(cont, 300, 220);
    lv_obj_center(cont);
    lv_obj_set_layout(cont, LV_LAYOUT_GRID);

    spinbox_pendulum_length = add_spinbox_row(g, cont, 0, "PendLen: ", -100, 100, 3, 0);
    spinbox_radius = add_spinbox_row(g, cont, 1, "Radius: ", -100, 100, 3, 0);
    spinbox_angle_in_degree = add_spinbox_row(g, cont, 2, "Angle: ", 0, 360, 3, 0);

    lv_spinbox_set_value(spinbox_radius, 15);
    lv_spinbox_set_value(spinbox_pendulum_length, 78);

    label_zero_info = lv_label_create(cont);
    lv_label_set_text(label_zero_info, "Bias: P=?,R=?");
    lv_obj_set_grid_cell(label_zero_info, LV_GRID_ALIGN_STRETCH, 0, 1, LV_GRID_ALIGN_CENTER, 3, 1);

    lv_obj_t *btn_zero = add_button_cell(g, cont, 3, 1, "Zero");
    lv_obj_add_event_cb(btn_zero, btn_zero_on_clicked, LV_EVENT_CLICKED, nullptr);

    lv_obj_t *btn_off = add_button_cell(g, cont, 4, 0, "Off");
    lv_obj_add_event_cb(btn_off, btn_off_on_clicked, LV_EVENT_CLICKED, nullptr);

    lv_obj_t *btn_brake = add_button_cell(g, cont, 4, 1, "Brake");
    lv_obj_add_event_cb(btn_brake, btn_brake_on_clicked, LV_EVENT_CLICKED, nullptr);

    lv_obj_t *btn_line = add_button_cell(g, cont, 5, 0, "Line");
    lv_obj_add_event_cb(btn_line, btn_line_on_clicked, LV_EVENT_CLICKED, nullptr);

    lv_obj_t *btn_circle = add_button_cell(g, cont, 5, 1, "Circle");
    lv_obj_add_event_cb(btn_circle, btn_circle_on_clicked, LV_EVENT_CLICKED, nullptr);

    label_imu_display = lv_label_create(cont);
    lv_label_set_text(label_imu_display, "IMU: P=?,R=?,Y=?");
    lv_obj_set_grid_cell(label_imu_display, LV_GRID_ALIGN_STRETCH, 0, 2, LV_GRID_ALIGN_CENTER, 6, 1);

    spinbox_pitch_p = add_spinbox_row(g, cont, 7, "PitchKp: ", 0, 99999999, 8, 5);
    spinbox_pitch_i = add_spinbox_row(g, cont, 8, "PitchKi: ", 0, 99999999, 8, 5);
    spinbox_pitch_d = add_spinbox_row(g, cont, 9, "PitchKd: ", 0, 99999999, 8, 5);
    spinbox_roll_p = add_spinbox_row(g, cont, 10, "RollKp: ", 0, 99999999, 8, 5);
    spinbox_roll_i = add_spinbox_row(g, cont, 11, "RollKi: ", 0, 99999999, 8, 5);
    spinbox_roll_d = add_spinbox_row(g, cont, 12, "RollKd: ", 0, 99999999, 8, 5);
    lv_spinbox_set_value(spinbox_pitch_p, static_cast<int32_t>(pid_pitch.Kp * 1e3));
    lv_spinbox_set_value(spinbox_pitch_i, static_cast<int32_t>(pid_pitch.Ki * 1e3));
    lv_spinbox_set_value(spinbox_pitch_d, static_cast<int32_t>(pid_pitch.Kd * 1e3));
    lv_spinbox_set_value(spinbox_roll_p, static_cast<int32_t>(pid_roll.Kp * 1e3));
    lv_spinbox_set_value(spinbox_roll_i, static_cast<int32_t>(pid_roll.Ki * 1e3));
    lv_spinbox_set_value(spinbox_roll_d, static_cast<int32_t>(pid_roll.Kd * 1e3));
    lv_obj_add_event_cb(spinbox_pitch_p, spinbox_pitch_p_on_changed, LV_EVENT_VALUE_CHANGED, nullptr);
    lv_obj_add_event_cb(spinbox_pitch_i, spinbox_pitch_i_on_changed, LV_EVENT_VALUE_CHANGED, nullptr);
    lv_obj_add_event_cb(spinbox_pitch_d, spinbox_pitch_d_on_changed, LV_EVENT_VALUE_CHANGED, nullptr);
    lv_obj_add_event_cb(spinbox_roll_p, spinbox_roll_p_on_changed, LV_EVENT_VALUE_CHANGED, nullptr);
    lv_obj_add_event_cb(spinbox_roll_i, spinbox_roll_i_on_changed, LV_EVENT_VALUE_CHANGED, nullptr);
    lv_obj_add_event_cb(spinbox_roll_d, spinbox_roll_d_on_changed, LV_EVENT_VALUE_CHANGED, nullptr);

    lv_timer_create(timer_update_imu_display, 200, nullptr);
}